__title__ = "MindsDB OpenBB handler"
__package_name__ = "mindsdb_openbb_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for the OpenBB Platform"
__author__ = "Didier Rodrigues Lopes"
__github__ = "https://github.com/openbb-finance/openbbterminal"
__pypi__ = "https://pypi.org/project/openbb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - OpenBB"
